#!/usr/bin/env python

import os
import os.path
import sys
import re
import shutil
import subprocess
import time
from xml.sax.saxutils import escape
import fnmatch

invokeDir = os.getcwd();
testsRoot = os.path.dirname(os.path.abspath(__file__))
if not os.path.exists(os.path.join(testsRoot, "tester.py")): raise "Can't find tests root directory!"
os.environ['OMK_TESTSROOT'] = testsRoot

class Rules(dict):
    "List of all rules and their snippets"
    def __init__(self):
        for r in os.listdir("../rules"):
            snippets = {}
            fn = os.path.join("..", "rules", r, "Makefile.rules")
            f = open(fn)
            for line in f:
                m = re.search("#OMK:(.*)\.omk", line);
                if m:
                    snippets[m.group(1)] = 1;
            self[r] = snippets.keys()
        print self

rules = Rules()

class Results(dict):
    def __init__(self):
        self.time = time.gmtime()
        self.datetime = time.strftime("%Y-%m-%d %H:%M:%S +0000", self.time)
        self.filename = "results-"+time.strftime("%Y%m%d-%H%M%S", self.time)+".html"
        self.stats = None

    def genStats(self):
        self.stats = Stats(self)
        
    def toHtml(self):
        s="""
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE html
  PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
  <meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1" />
  <title>OMK test report %(datetime)s</title>
</head>
<body>
<h2>Summary</h2>
        """ % self.__dict__
        s+=self.stats.toHtml()
        s+="""
<h2>Chart by rules</h2>
        """
        s+=self.toHtmlByRules();
        s+="""
<h2>Chart by testcase</h2>
<table cellpadding='2' border='1'>
<tbody>
        """
        tests = sorted(self.keys())
        for t in tests:
            s+=self[t].toHtml()
        s+="""
</tbody></table>
<h2>Outputs of tests</h2>
        """
        for t in tests:
            s+=self[t].toHtmlOutputs()
        s+="""
</body></html>
        """
        return s

    def toHtmlByRules(self):
        rk = sorted(rules.keys())
        tests = sorted(self.keys())
        s=''
        for r in rk:
            s+="""
<a name='results-%(rules)s' />
<h3>Rules: %(rules)s</h3>
<table cellpadding='2' border='1'>
<tbody>
""" % { 'rules':r }
            for t in tests:
                if self[t].has_key(r):
                    s+=self[t][r].toHtml("<a href='%s'>/</a><a href='%s'>%s</a>" %
                                         (self[t].tc.dirRelative, self[t].tc.scriptNameRelative, self[t].tc.name))
            s+="""
</tbody></table>
"""
        return s

    def save(self):
        f = file(self.filename, "w+")
        f.write(self.toHtml())
        try:
            os.remove("results.html")
        except:
            pass
        os.symlink(self.filename, "results.html")
        print "Results written to "+self.filename

class TestCaseResult(dict):
    def __init__(self, tc):
        self.tc = tc
        
    def toHtml(self):
        rules = sorted(self.keys())
        s="""
  <tr><td colspan='5'><strong><a href='%s'>%s</a></strong></td></tr>
        """ % (self.tc.scriptNameRelative, self.tc.name)
        for r in rules:
            s+=self[r].toHtml(r)
        return s

    def toHtmlOutputs(self):
        rules = sorted(self.keys())
        s="<h3>Testcase: %s</h3>" % self.tc.name
        for r in rules:
            s+=self[r].toHtmlOutputs()
        return s

class ResultEntry:
    def __init__(self, tcname, rules):
        self.tcname = tcname
        self.rules = rules
        self.canttest = 0
        
    def toHtml(self, title):
        if self.exitcode == 0: color=''
        elif self.canttest: color=' bgcolor="gold"'
        else: color=' bgcolor="red"'
        if self.stdout: stdoutlink="<a href='#output-%(tcname)s-%(rules)s'>output</a>" % self.__dict__
        else: stdoutlink=''
        s="""
  <tr%(color)s>
    <td>%(title)s</td>
    <td>%(exitcode)d</td>
    <td>%(message)s</td>
    <td>%(time).1f s</td>
    <td>%(stdoutlink)s</td>
  </tr>
        """ % {
            'color' : color,
            'tcname' : self.rules,
            'title' : title,
            'exitcode' : self.exitcode,
            'message' : escape(self.message),
            'time' : self.time,
            'stdoutlink' : stdoutlink,
            }
        return s
    def toHtmlOutputs(self):
        vals = {
            'tcname':self.tcname,
            'rules':self.rules,
            'stdout':escape(self.stdout),
            }
        s=""
        if self.stdout: s+="""
<a name='output-%(tcname)s-%(rules)s'/>
<h5>Output of test: %(tcname)s, rules: %(rules)s</h5>
<pre>%(stdout)s</pre>""" % vals
        return s

class RulesStat:
    def __init__(self, rules):
        self.rules = rules
        self.tests = 0
        self.success = 0
        self.errors = 0
        self.canttest = 0
    def update(self, testCaseResult):
        try:
            resultEntry = testCaseResult[self.rules]
            self.tests+=1
            if resultEntry.exitcode == 0: self.success+=1
            elif resultEntry.canttest: self.canttest+=1
            else: self.errors+=1
        except KeyError:
            pass
    def toHtml(self):
        if self.errors == 0 and self.canttest == 0: self.color=''
        elif self.errors != 0:   self.color=' bgcolor="red"'
        else: self.color = ' bgcolor="gold"'
        s="""
  <tr%(color)s>
    <td><a href='#results-%(rules)s'>%(rules)s</a></td>
    <td>%(tests)d</td>
    <td>%(success)d</td>
    <td>%(errors)d</td>
    <td>%(canttest)d</td>
  </tr>
        """ % self.__dict__
        return s

class Stats(dict):
    def __init__(self, results):
        rk = rules.keys()
        for rule in rk:
            rulesStat = RulesStat(rule)
            self[rule]=rulesStat
            for resultEntry in results.values():
                rulesStat.update(resultEntry)
        
    def toHtml(self):
        s="""
<table cellpadding='2' border='1'>
<col />
<col span='5' align='right' />
<thead><tr>
  <td>Rules</td>
  <td>Total</td>
  <td>Success</td>
  <td>Errors</td>
  <td>Can't test</td>
</tr></thead>
<tbody>
        """
        rules = sorted(self.keys())
        for r in rules:
            s+=self[r].toHtml()
        s+="""
</tbody></table>"""
        return s;

class TestCase:
    def __init__(self, directory, executable):
        self.directory = directory      # Absolute directory
        self.executable = executable
        self.name = self._getName()
        self.scriptNameRelative = self._getScriptRelative();
        self.dirRelative = os.path.dirname(self.scriptNameRelative)
        self._whichRules()

    def _getName(self):
        name = self.directory
        if name.startswith(testsRoot+"/"):
            name = name[len(testsRoot)+1:]
        testSuffix = re.match("^runtest[-_. :]*(.*)", self.executable).group(1)
        if testSuffix:
            name+=" "+testSuffix
        return name

    def _getScriptRelative(self):
        script = os.path.join(self.directory, self.executable);
        if script.startswith(invokeDir+"/"):
            script = script[len(invokeDir)+1:]
        return script

    def _whichRules(self):
        """Reads the rules file and creates the self.rules list of all
        rules to test"""
        self.rules = []
        try:
            f = open(os.path.join(self.directory, self.executable+'.rules'))
        except IOError:
            self.rules = rules.keys()
            return
        line = f.readline()
        colonMatch = re.search('([^:]*) *: *(.*)', line)
        if colonMatch:
            if colonMatch.group(1) == "all":
                # all:
                self.rules = rules.keys()
            elif colonMatch.group(1) == "snip":
                # snip: ...
                snip = colonMatch.group(2)
                for r in rules:
                    if snip in rules[r]:
                        self.rules.append(r)
            elif colonMatch.group(1) == "python":
                # python: ...
                expr = colonMatch.group(2)
                for r in rules:
                    if eval(expr, {'rules': r, 'snippets': rules[r]}):
                        self.rules.append(r)
        else:
            # rule name
            line = line.strip()
            if line in rules: self.rules = [ line ]
        #print self.rules
        

    def run(self):
        self.results = TestCaseResult(self)
        print "Testing \"%s\" in %s:\n" % (self.name, self.executable),
        os.chdir(os.path.join(testsRoot, self.directory))
        # Chose how to run the test - currently there is only one option
#         if os.path.exists("Makefile.test"):
#             self._exec = self._execMake
        if os.path.exists(self.executable):
            self._exec = self._execRuntest
        else: return
        self.failed = False
        for rules in self.rules:
            resultEntry = ResultEntry(self.name, rules)
            self.results[rules] = resultEntry
            os.environ['OMK_RULES'] = rules
            filesBefore = self._getFileSet()
            self._copyRules(rules)
            try:
                self._doRun(resultEntry)
            finally:
                filesAfter = self._getFileSet()
                self._clean(filesBefore, filesAfter)
        print

    def _getFileSet(self):
        files = set()
        for f in os.listdir("."):
            files.add(f)
        return files
        
    def _clean(self, filesBefore, filesAfter):
        remove = filesAfter - filesBefore
        for f in remove:
            os.system("rm -rf "+f)

#     def _execMake(self):
#         return os.system("make -k -f Makefile.test > /dev/null 2>&1")

    def _execRuntest(self, log):
        startTime = time.clock()
        pipe = subprocess.Popen("./"+self.executable, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, close_fds=True)
        (output, error) = pipe.communicate()
        endTime = time.clock()
        ret = pipe.returncode
        log.time = endTime - startTime
        log.exitcode = ret
        log.stdout = output
        log.message = ''
        if os.path.exists("_canttest"):
            log.message = file("_canttest").read()
            log.canttest = 1
            ret = 2                     # If Makefile sets canttest, override the exit code which might be zero
            log.exitcode = ret
        elif ret != 0:
            if os.path.exists("_error"):
                log.message = file("_error").read()
        return ret

    def _copyRules(self, rules):
        "Copies the rules to the current directory"
        src = os.path.join(testsRoot, "..", "rules", rules, "Makefile.rules")
        shutil.copy(src, ".")
    
    def _doRun(self, log):
        "Runs the teset in current directory."
#         print "    ",os.environ['OMK_RULES'],
#         sys.stdout.flush()
        ret = self._exec(log)
#         if log.canttest: retstr = "--"
#         elif ret == 0: retstr = "ok"
#         else:
#             retstr = "FAILED"
#             self.failed = True
#         print "%*s%s" % (20-len(os.environ['OMK_RULES']), "", retstr)
        if log.canttest: retstr = "[not tested] ok"
        elif ret == 0: retstr = "ok"
        else:
            retstr = "FAILED"
            self.failed = True
        print "! %s (%s) %s" % (self.name, os.environ['OMK_RULES'], retstr)
        sys.stdout.flush()


results = Results()

failed = False;
for dirpath, dirnames, filenames in os.walk(invokeDir):
    executables = fnmatch.filter(filenames, "runtest*")
    if not executables: continue
    for exe in executables:
	if exe[-1] == "~": continue
        if re.search(".rules$", exe): continue
        t = TestCase(dirpath, exe)
        t.run()
        if t.failed: failed = True
        results[t.name] = t.results

os.chdir(invokeDir)
results.genStats()
results.save()

sys.exit(failed)

# Local Variables:
# compile-command: "python tester.py"
# End:
